@@ -1,3 +1,4 @@
+import glob
 import imghdr
 import os
 import tempfile
@@ -13,11 +14,29 @@ from PIL import Image
 from robot.libraries.BuiltIn import BuiltIn
 
 from RPA.PDF.keywords import LibraryContext, keyword
+from RPA.core.robocorp import robocorp_home
 from .model import Document, Figure
 
 
 ListOrString = Union[List[int], List[str], str, None]
 
+ASSETS_DIR = Path(__file__).resolve().parent.parent / "assets"
+
+
+def get_output_dir() -> Path:
+    try:
+        # A `None` may come from here too.
+        output_dir = BuiltIn().get_variable_value("${OUTPUT_DIR}")
+    except Exception:  # pylint: disable=broad-except
+        output_dir = None
+    # Keep empty string as current working directory path.
+    if output_dir is None:
+        output_dir = "output"
+
+    output_dir = Path(output_dir).expanduser().resolve()
+    output_dir.mkdir(parents=True, exist_ok=True)
+    return output_dir
+
 
 class PDF(FPDF, HTMLMixin):
     """
@@ -27,25 +46,51 @@ class PDF(FPDF, HTMLMixin):
     https://github.com/PyFPDF/fpdf2
     """
 
+    FONT_PATHS = {
+        "": ASSETS_DIR / "Inter-Regular.ttf",
+        "B": ASSETS_DIR / "Inter-Bold.ttf",
+        "I": ASSETS_DIR / "Inter-Italic.ttf",
+        "BI": ASSETS_DIR / "Inter-BoldItalic.ttf",
+    }
+    FONT_CACHE_DIR = robocorp_home() / "fonts"
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.font_cache_dir = self.FONT_CACHE_DIR
+        self.font_cache_dir.mkdir(parents=True, exist_ok=True)
+
+    # pylint: disable=arguments-differ
+    def add_font(self, *args, fname, **kwargs):
+        try:
+            return super().add_font(*args, fname=fname, **kwargs)
+        # pylint: disable=broad-except
+        except Exception:
+            # Usually caching issues, like importing a *.pkl font file serialized on
+            # another OS/env.
+            unifilename = self.font_cache_dir / f"{fname.stem}.pkl"
+            if unifilename.exists():
+                os.remove(unifilename)
+            return super().add_font(*args, fname=fname, **kwargs)
+
+    def add_unicode_fonts(self):
+        for style, path in self.FONT_PATHS.items():
+            self.add_font("Inter", style=style, fname=path, uni=True)
+        self.set_font("Inter")
+
 
 class DocumentKeywords(LibraryContext):
     """Keywords for basic PDF operations"""
 
-    def resolve_output(self, path: Optional[str] = None) -> str:
+    ENCODING = "utf-8"
+
+    @staticmethod
+    def resolve_output(path: Optional[str] = None) -> str:
         if path is None:
-            try:
-                output_dir = BuiltIn().get_variable_value("${OUTPUT_DIR}")
-            except Exception:  # pylint: disable=broad-except
-                output_dir = None
-
-            if output_dir is None:
-                output_dir = "output"
-
-            output = Path(output_dir) / "output.pdf"
-            output = output.resolve()
+            output = get_output_dir() / "output.pdf"
         else:
-            output = Path(path).resolve()
+            output = Path(path)
 
+        output = output.expanduser().resolve()
         output.parent.mkdir(parents=True, exist_ok=True)
         return str(output)
 
@@ -133,6 +178,7 @@ class DocumentKeywords(LibraryContext):
         template: str,
         output_path: str,
         variables: dict = None,
+        encoding: str = ENCODING,
     ) -> None:
         """Use HTML template file to generate PDF file.
 
@@ -176,24 +222,26 @@ class DocumentKeywords(LibraryContext):
             }
             p.template_html_to_pdf("order.template", "order.pdf", data)
 
-        :param template: filepath to the HTML template.
-        :param output_path: filepath where to save PDF document.
-        :param variables: dictionary of variables to fill into template, defaults to {}.
+        :param template: Filepath to the HTML template.
+        :param output_path: Filepath where to save PDF document.
+        :param variables: Dictionary of variables to fill into template, defaults to {}.
+        :param encoding: Codec used for text I/O.
         """
         variables = variables or {}
 
-        with open(template, "r", encoding="utf-8") as templatefile:
+        with open(template, "r", encoding=encoding or self.ENCODING) as templatefile:
             html = templatefile.read()
         for key, value in variables.items():
             html = html.replace("{{" + key + "}}", str(value))
 
-        self.html_to_pdf(html, output_path)
+        self.html_to_pdf(html, output_path, encoding=encoding)
 
     @keyword
     def html_to_pdf(
         self,
         content: str,
         output_path: str,
+        encoding: str = ENCODING,
     ) -> None:
         """Generate a PDF file from HTML content.
 
@@ -222,17 +270,29 @@ class DocumentKeywords(LibraryContext):
                 pdf.html_to_pdf(html_content_as_string, "/tmp/output.pdf")
 
         :param content: HTML content.
-        :param output_path: filepath where to save the PDF document.
+        :param output_path: Filepath where to save the PDF document.
+        :param encoding: Codec used for text I/O.
         """
         output_path = self.resolve_output(output_path)
         self.logger.info("Writing output to file %s", output_path)
 
-        fpdf = PDF()
-        fpdf.set_margin(0)
-        fpdf.add_page()
-        fpdf.write_html(content)
-        fpdf.output(name=output_path)
-        fpdf = PDF()
+        def _html_to_pdf():
+            fpdf = PDF()
+            # Support unicode content with a font capable of rendering it.
+            fpdf.core_fonts_encoding = encoding
+            fpdf.add_unicode_fonts()
+            fpdf.set_margin(0)
+            fpdf.add_page()
+            fpdf.write_html(content)
+            fpdf.output(name=output_path)
+
+        try:
+            _html_to_pdf()
+        except FileNotFoundError:
+            serialized_fonts = glob.glob(str(PDF.FONT_CACHE_DIR / "*.pkl"))
+            for serialized_font in serialized_fonts:
+                os.remove(serialized_font)
+            _html_to_pdf()
 
     @keyword
     def get_pdf_info(self, source_path: str = None) -> dict:
@@ -269,11 +329,11 @@ class DocumentKeywords(LibraryContext):
         """
         self.switch_to_pdf(source_path)
 
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         docinfo = reader.getDocumentInfo()
         num_pages = reader.getNumPages()
 
-        parser = PDFParser(self.ctx.active_pdf_document.fileobject)
+        parser = PDFParser(self.active_pdf_document.fileobject)
         document = PDFDocument(parser)
         try:
             fields = pdfminer.pdftypes.resolve1(document.catalog["AcroForm"])["Fields"]
@@ -328,7 +388,7 @@ class DocumentKeywords(LibraryContext):
                 is_encrypted = pdf.is_pdf_encrypted("/tmp/sample.pdf")
         """
         self.switch_to_pdf(source_path)
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         return reader.isEncrypted
 
     @keyword
@@ -365,7 +425,7 @@ class DocumentKeywords(LibraryContext):
         :raises PdfReadError: if file is encrypted or other restrictions are in place
         """
         self.switch_to_pdf(source_path)
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         return reader.getNumPages()
 
     @keyword
@@ -405,19 +465,19 @@ class DocumentKeywords(LibraryContext):
             file to activate.
         """
         if not source_path:
-            if not self.ctx.active_pdf_document:
+            if not self.active_pdf_document:
                 raise ValueError("No PDF is open")
             self.logger.debug(
-                "Using already set document: %s", self.ctx.active_pdf_document.path
+                "Using already set document: %s", self.active_pdf_document.path
             )
             return
 
         source_path = str(source_path)
         if source_path not in self.ctx.documents:
             self.open_pdf(source_path)
-        elif self.ctx.documents[source_path] != self.ctx.active_pdf_document:
+        elif self.ctx.documents[source_path] != self.active_pdf_document:
             self.logger.debug("Switching to already opened document: %s", source_path)
-            self.ctx.active_pdf_document = self.ctx.documents[source_path]
+            self.active_pdf_document = self.ctx.documents[source_path]
         else:
             self.logger.debug("Using already set document: %s", source_path)
 
@@ -468,14 +528,14 @@ class DocumentKeywords(LibraryContext):
         self.switch_to_pdf(source_path)
         self.ctx.convert(trim=trim)
 
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         pages = self._get_page_numbers(pages, reader)
         pdf_text = {}
         for idx, page in self.active_pdf_document.get_pages().items():
             if page.pageid not in pages:
                 continue
             pdf_text[idx] = [] if details else ""
-            for _, item in page.get_textboxes().items():
+            for _, item in page.textboxes.items():
                 if details:
                     pdf_text[idx].append(item)
                 else:
@@ -533,7 +593,7 @@ class DocumentKeywords(LibraryContext):
             if None then extracts all pages.
         """
         self.switch_to_pdf(source_path)
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         writer = PyPDF2.PdfFileWriter()
 
         output_path = self.resolve_output(output_path)
@@ -597,7 +657,7 @@ class DocumentKeywords(LibraryContext):
         """
         # TODO: don't save to a new file every time
         self.switch_to_pdf(source_path)
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         writer = PyPDF2.PdfFileWriter()
 
         output_path = self.resolve_output(output_path)
@@ -664,7 +724,7 @@ class DocumentKeywords(LibraryContext):
         """
         # TODO: don't save to a new file every time
         self.switch_to_pdf(source_path)
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
 
         output_path = self.resolve_output(output_path)
 
@@ -713,7 +773,7 @@ class DocumentKeywords(LibraryContext):
         :raises ValueError: on decryption errors.
         """
         self.switch_to_pdf(source_path)
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         try:
             match_result = reader.decrypt(password)
 
@@ -775,7 +835,7 @@ class DocumentKeywords(LibraryContext):
         self.ctx.convert()
         pages = {}
         for pagenum, page in self.active_pdf_document.get_pages().items():
-            pages[pagenum] = page.get_figures()
+            pages[pagenum] = page.figures
         return pages
 
     @keyword
@@ -833,7 +893,7 @@ class DocumentKeywords(LibraryContext):
         writer = PyPDF2.PdfFileWriter()
         pdf = FPDF()
         pdf.add_page()
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         mediabox = reader.getPage(0).mediaBox
         im = Image.open(image_path)
         max_width = int(float(mediabox.getWidth()) * coverage)
