@@ -4,8 +4,11 @@ import typing
 from collections import OrderedDict
 from typing import (
     Any,
+    Set,
     Iterable,
     Optional,
+    Tuple,
+    Union,
 )
 
 import PyPDF2
@@ -39,135 +42,133 @@ from RPA.PDF.keywords import (
 )
 
 
-def iterable_items_to_int(bbox) -> list:
+Coords = Tuple[int, ...]
+
+
+def iterable_items_to_ints(bbox: Optional[Iterable]) -> Coords:
     if bbox is None:
-        return []
-    return list(map(int, bbox))
+        return ()
+    return tuple(map(int, bbox))
 
 
-class Figure:
+class BaseElement:
+    """Base class for all kind of elements found in PDFs."""
+
+    def __init__(self, bbox: Optional[Iterable]):
+        self._bbox: Coords = iterable_items_to_ints(bbox)
+        assert len(self._bbox) == 4, "must be in (left, bottom, right, top) format"
+
+    @property
+    def bbox(self) -> Coords:
+        return self._bbox
+
+    @property
+    def left(self) -> int:
+        return self.bbox[0]
+
+    @property
+    def bottom(self) -> int:
+        return self.bbox[1]
+
+    @property
+    def right(self) -> int:
+        return self.bbox[2]
+
+    @property
+    def top(self) -> int:
+        return self.bbox[3]
+
+
+class Figure(BaseElement):
     """Class for each LTFigure element in the PDF"""
 
-    figure_name: str
-    figure_bbox: list
-    item: dict
-    image_name: str
+    def __init__(self, item):
+        super().__init__(item.bbox)
+        self._item = item
 
-    def __init__(self, name: str, bbox: Iterable) -> None:
-        self.figure_name = name
-        self.figure_bbox = iterable_items_to_int(bbox)
-        self.image_name = None
-        self.item = None
+    @property
+    def item(self):
+        return self._item
 
-    def set_item(self, item: Any):
-        # LTImage
-        self.item = item
-
-    def details(self) -> str:
-        return '<image src="%s" width="%d" height="%d" />' % (
-            self.image_name or self.figure_name,
-            self.item.width,
-            self.item.height,
+    def __str__(self) -> str:
+        return (
+            f'<image src="{self.item.name}" width="{int(self.item.width)}" '
+            f'height="{int(self.item.height)}" />'
         )
 
 
-class TextBox:
-    """Class for each LTTextBox element in the PDF"""
+class TextBox(BaseElement):
+    """Class for each LTTextBox element in the PDF."""
 
-    item: dict
-    textbox_bbox: list
-    textbox_id: int
-    textbox_wmode: str
+    def __init__(self, boxid: int, *, item: Any, trim: bool = True) -> None:
+        super().__init__(item.bbox)
 
-    def __init__(
-        self, boxid: int, bbox: Iterable, wmode: str, trim: bool = True
-    ) -> None:
-        self.textbox_id = boxid
-        self.textbox_bbox = iterable_items_to_int(bbox)
-        self.textbox_wmode = wmode
-        self.trim = trim
-
-    def set_item(self, item: Any):
-        text = item.get_text()
-        self.item = {
-            "bbox": iterable_items_to_int(item.bbox),
-            "text": text.strip() if self.trim else text,
-        }
-
-    @property
-    def left(self) -> Any:
-        return self.bbox[0] if (self.bbox and len(self.bbox) == 4) else None
-
-    @property
-    def bottom(self) -> Any:
-        return self.bbox[1] if (self.bbox and len(self.bbox) == 4) else None
-
-    @property
-    def right(self) -> Any:
-        return self.bbox[2] if (self.bbox and len(self.bbox) == 4) else None
-
-    @property
-    def top(self) -> Any:
-        return self.bbox[3] if (self.bbox and len(self.bbox) == 4) else None
+        self._boxid = boxid
+        self._text = item.get_text()
+        if trim:
+            self._text = self._text.strip()
 
     @property
     def boxid(self) -> int:
-        return self.textbox_id
+        return self._boxid
 
     @property
     def text(self) -> str:
-        return self.item["text"]
-
-    @text.setter
-    def text(self, newtext):
-        self.item["text"] = newtext
-
-    @property
-    def bbox(self) -> list:
-        return self.item["bbox"]
+        return self._text
 
     def __str__(self) -> str:
         return f"{self.text} {self.bbox}"
 
 
-class Page:
-    """Class for each PDF page"""
-
-    bbox: list
-    content: OrderedDict
-    content_id: int
-    pageid: str
-    rotate: int
+class Page(BaseElement):
+    """Class that abstracts a PDF page."""
 
     def __init__(self, pageid: int, bbox: Iterable, rotate: int) -> None:
+        super().__init__(bbox)
+
         self.pageid = pageid
-        self.bbox = iterable_items_to_int(bbox)
         self.rotate = rotate
-        self.content = OrderedDict()
-        self.content_id = 0
+
+        self._content = OrderedDict()
+        self._content_id = 0
+        self._figures = OrderedDict()
+        self._textboxes = OrderedDict()
 
     def add_content(self, content: Any) -> None:
-        self.content[self.content_id] = content
-        self.content_id += 1
+        self._content[self._content_id] = content
+        if isinstance(content, Figure):
+            content_dict = self._figures
+        elif isinstance(content, TextBox):
+            content_dict = self._textboxes
+        else:
+            content_dict = None
+        if content_dict is not None:
+            content_dict[self._content_id] = content
 
-    def get_content(self) -> OrderedDict:
-        return self.content
+        self._content_id += 1
 
-    def get_figures(self) -> OrderedDict:
-        return {k: v for k, v in self.content.items() if isinstance(v, Figure)}
+    @property
+    def content(self) -> OrderedDict:
+        return self._content
 
-    def get_textboxes(self) -> OrderedDict:
-        return {k: v for k, v in self.content.items() if isinstance(v, TextBox)}
+    @property
+    def figures(self) -> OrderedDict:
+        return self._figures
+
+    @property
+    def textboxes(self) -> OrderedDict:
+        return self._textboxes
+
+    @property
+    def tag(self) -> str:
+        return (
+            f'<page id="{self.pageid}" bbox="{bbox2str(self.bbox)}" '
+            f'rotate="{self.rotate}">'
+        )
 
     def __str__(self) -> str:
-        page_as_str = '<page id="%s" bbox="%s" rotate="%d">\n' % (
-            self.pageid,
-            bbox2str(self.bbox),
-            self.rotate,
-        )
-        for _, c in self.content.items():
-            page_as_str += f"{c}\n"
-        return page_as_str
+        items_str = "\n".join(self._content.values())
+        return f"{self.tag}\n{items_str}"
 
 
 class Document:
@@ -182,7 +183,7 @@ class Document:
         self._pages = OrderedDict()
         self._xml_content_list: typing.List[bytes] = []
         self.fields: Optional[dict] = None
-        self.is_converted = False
+        self.has_converted_pages: Set[int] = set()
 
     @property
     def path(self):
@@ -226,6 +227,8 @@ class Converter(PDFConverter):
         self,
         active_document: Document,
         rsrcmgr,
+        *,
+        logger,
         codec: str = "utf-8",
         pageno: int = 1,
         laparams=None,
@@ -237,13 +240,21 @@ class Converter(PDFConverter):
             rsrcmgr, sys.stdout, codec=codec, pageno=pageno, laparams=laparams
         )
         self.active_pdf_document = active_document
-        self.figure = None
         self.current_page = None
         self.imagewriter = imagewriter
         self.stripcontrol = stripcontrol
         self.trim = trim
         self.write_header()
 
+        self._logger = logger
+        self._unique_figures: Set[Tuple[int, str, Coords]] = set()
+
+    def _add_unique_figure(self, figure: Figure):
+        figure_key = (self.current_page.pageid, str(figure), figure.bbox)
+        if figure_key not in self._unique_figures:
+            self.current_page.add_content(figure)
+            self._unique_figures.add(figure_key)
+
     def write(self, text: str):
         if self.codec:
             text = text.encode(self.codec)
@@ -283,14 +294,8 @@ class Converter(PDFConverter):
         #  pylint: disable=R0912, R0915
         def render(item):
             if isinstance(item, LTPage):
-                s = '<page id="%s" bbox="%s" rotate="%d">\n' % (
-                    item.pageid,
-                    bbox2str(item.bbox),
-                    item.rotate,
-                )
                 self.current_page = Page(item.pageid, item.bbox, item.rotate)
-
-                self.write(s)
+                self.write(self.current_page.tag + "\n")
                 for child in item:
                     render(child)
                 if item.groups is not None:
@@ -320,20 +325,16 @@ class Converter(PDFConverter):
                 )
                 self.write(s)
             elif isinstance(item, LTFigure):
-                self.figure = Figure(item.name, item.bbox)
-                if self.figure:
-                    s = '<figure name="%s" bbox="%s">\n' % (
-                        item.name,
-                        bbox2str(item.bbox),
-                    )
-                    self.write(s)
-                    for child in item:
-                        if self.figure:
-                            self.figure.set_item(item)
-                        render(child)
-                    self.write("</figure>\n")
-                    self.current_page.add_content(self.figure)
-                    self.figure = None
+                figure = Figure(item)
+                s = '<figure name="%s" bbox="%s">\n' % (
+                    item.name,
+                    bbox2str(item.bbox),
+                )
+                self.write(s)
+                for child in item:
+                    render(child)
+                self.write("</figure>\n")
+                self._add_unique_figure(figure)
             elif isinstance(item, LTTextLine):
                 self.write('<textline bbox="%s">\n' % bbox2str(item.bbox))
                 for child in item:
@@ -349,9 +350,8 @@ class Converter(PDFConverter):
                     bbox2str(item.bbox),
                     wmode,
                 )
-                box = TextBox(item.index, item.bbox, wmode, self.trim)
+                box = TextBox(item.index, item=item, trim=self.trim)
                 self.write(s)
-                box.set_item(item)
                 self.current_page.add_content(box)
                 for child in item:
                     render(child)
@@ -374,8 +374,7 @@ class Converter(PDFConverter):
             elif isinstance(item, LTText):
                 self.write("<text>%s</text>\n" % item.get_text())
             elif isinstance(item, LTImage):
-                if self.figure:
-                    self.figure.set_item(item)
+                figure = Figure(item)
                 if self.imagewriter is not None:
                     name = self.imagewriter.export_image(item)
                     self.write(
@@ -386,8 +385,9 @@ class Converter(PDFConverter):
                     self.write(
                         '<image width="%d" height="%d" />\n' % (item.width, item.height)
                     )
+                self._add_unique_figure(figure)
             else:
-                assert False, str(("Unhandled", item))
+                self._logger.warning("Unknown item: %r", item)
 
         render(ltpage)
 
@@ -398,8 +398,15 @@ class Converter(PDFConverter):
 class ModelKeywords(LibraryContext):
     """Keywords for converting PDF document into specific RPA object model"""
 
+    FIELDS_ENCODING = "iso-8859-1"
+
     @keyword
-    def convert(self, source_path: str = None, trim: bool = True) -> None:
+    def convert(
+        self,
+        source_path: str = None,
+        trim: bool = True,
+        pagenum: Optional[Union[int, str]] = None,
+    ):
         """Parse source PDF into entities.
 
         These entities can be used for text searches or XML dumping for example. The
@@ -408,6 +415,8 @@ class ModelKeywords(LibraryContext):
 
         :param source_path: source PDF filepath
         :param trim: trim whitespace from the text is set to True (default)
+        :param pagenum: Page number where search is performed on, defaults to `None`. (
+            meaning all pages get converted)
 
         **Examples**
 
@@ -434,43 +443,86 @@ class ModelKeywords(LibraryContext):
                 pdf.convert("/tmp/sample.pdf")
         """
         self.ctx.switch_to_pdf(source_path)
-        if self.ctx.active_pdf_document.is_converted:
-            return
+        converted_pages = self.active_pdf_document.has_converted_pages
+        if pagenum is not None:
+            pagenum = int(pagenum)
+            if pagenum in converted_pages:
+                return  # specific page already converted
+        else:
+            pages_count = self.active_pdf_document.reader.getNumPages()
+            if len(converted_pages) >= pages_count:
+                return  # all pages got converted already
 
         self.logger.debug(
-            "Converting active PDF document: %s", self.ctx.active_pdf_document.path
+            "Converting active PDF document page %s on: %s",
+            pagenum if pagenum is not None else "<all>",
+            self.active_pdf_document.path,
         )
         rsrcmgr = PDFResourceManager()
         if not self.ctx.convert_settings:
             self.set_convert_settings()
         laparams = pdfminer.layout.LAParams(**self.ctx.convert_settings)
         device = Converter(
-            self.ctx.active_pdf_document, rsrcmgr, laparams=laparams, trim=trim
+            self.active_pdf_document,
+            rsrcmgr,
+            laparams=laparams,
+            trim=trim,
+            logger=self.logger,
+            # Also explicitly set by us when iterating pages for processing.
+            pageno=pagenum if pagenum is not None else 1,
         )
         interpreter = pdfminer.pdfinterp.PDFPageInterpreter(rsrcmgr, device)
 
         # Look at all (nested) objects on each page.
-        source_parser = PDFParser(self.ctx.active_pdf_document.fileobject)
+        source_parser = PDFParser(self.active_pdf_document.fileobject)
         source_document = PDFDocument(source_parser)
         source_pages = PDFPage.create_pages(source_document)
-        for _, page in enumerate(source_pages, 0):
-            interpreter.process_page(page)
+        for idx, page in enumerate(source_pages, start=1):
+            # Process relevant pages only if instructed like so.
+            # (`pagenum` starts from 1 as well)
+            if pagenum is None or idx == pagenum:
+                if idx not in converted_pages:
+                    # Skipping converted pages will leave this counter un-incremented,
+                    # therefore we increment it explicitly.
+                    device.pageno = idx
+                    interpreter.process_page(page)
+                    converted_pages.add(idx)
+
         device.close()
 
-        self.ctx.active_pdf_document.is_converted = True
+    @classmethod
+    def _decode_field(
+        cls, binary: Optional[bytes], *, encoding
+    ) -> Optional[Union[str, bytes]]:
+        if not (binary and hasattr(binary, "decode")):
+            return binary
+
+        try:
+            return binary.decode(encoding)
+        except UnicodeDecodeError:
+            return binary.decode(cls.FIELDS_ENCODING)
 
     @keyword
     def get_input_fields(
-        self, source_path: str = None, replace_none_value: bool = False
+        self,
+        source_path: Optional[str] = None,
+        replace_none_value: bool = False,
+        encoding: str = FIELDS_ENCODING,
     ) -> dict:
         """Get input fields in the PDF.
 
-        Stores input fields internally so that they can be used without
-        parsing the PDF again.
+        Stores input fields internally so that they can be used without parsing the PDF
+        again.
 
-        Parameter ``replace_none_value`` is for convience to visualize fields.
-
-        If no source path given, assumes a PDF is already opened.
+        :param source_path: Filepath to source, if not given use the currently active
+            PDF.
+        :param replace_none_value: Enable this to conveniently visualize the fields. (
+            replaces the null value with field's name)
+        :param encoding: Use an explicit encoding for field name/value parsing. (
+            defaults to "iso-8859-1" but "utf-16" might work for you)
+        :returns: A dictionary with all the found fields. Use their key names when
+            setting values into them.
+        :raises KeyError: If no input fields are enabled in the PDF.
 
         **Examples**
 
@@ -478,12 +530,9 @@ class ModelKeywords(LibraryContext):
 
         .. code-block:: robotframework
 
-            ***Settings***
-            Library    RPA.PDF
-
-            ***Tasks***
             Example Keyword
-                ${fields}=  Get Input Fields    /tmp/sample.pdf
+                ${fields} =     Get Input Fields    form.pdf
+                Log Dictionary    ${fields}
 
         **Python**
 
@@ -494,15 +543,13 @@ class ModelKeywords(LibraryContext):
             pdf = PDF()
 
             def example_keyword():
-                fields = pdf.get_input_fields("/tmp/sample.pdf")
+                fields = pdf.get_input_fields("form.pdf")
+                print(fields)
 
-        :param source_path: source filepath, defaults to None.
-        :param replace_none_value: if value is None replace it with key name,
-            defaults to False.
-        :return: dictionary of input key values or `None`.
+            example_keyword()
         """
         self.ctx.switch_to_pdf(source_path)
-        active_document = self.ctx.active_pdf_document
+        active_document = self.active_pdf_document
 
         active_fields = active_document.fields
         if active_fields:
@@ -518,7 +565,7 @@ class ModelKeywords(LibraryContext):
         except KeyError as err:
             raise KeyError(
                 'PDF "%s" does not have any input fields.'
-                % self.ctx.active_pdf_document.path
+                % self.active_pdf_document.path
             ) from err
 
         record_fields = {}
@@ -527,34 +574,22 @@ class ModelKeywords(LibraryContext):
             if field is None:
                 continue
 
-            name, value, rect, label = (
-                field.get("T"),
-                field.get("V"),
+            name, value, raw_rect, label = (
+                self._decode_field(field.get("T"), encoding=encoding),
+                self._decode_field(field.get("V"), encoding=encoding),
                 field.get("Rect"),
-                field.get("TU"),
+                self._decode_field(field.get("TU"), encoding=encoding),
             )
             if value is None and replace_none_value:
-                record_fields[name.decode("iso-8859-1")] = {
-                    "value": name.decode("iso-8859-1"),
-                    "rect": iterable_items_to_int(rect),
-                    "label": label.decode("iso-8859-1") if label else None,
-                }
-            else:
-                try:
-                    record_fields[name.decode("iso-8859-1")] = {
-                        "value": value.decode("iso-8859-1") if value else "",
-                        "rect": iterable_items_to_int(rect),
-                        "label": label.decode("iso-8859-1") if label else None,
-                    }
-                except AttributeError:
-                    self.logger.debug("Attribute error")
-                    record_fields[name.decode("iso-8859-1")] = {
-                        "value": value,
-                        "rect": iterable_items_to_int(rect),
-                        "label": label.decode("iso-8859-1") if label else None,
-                    }
+                value = name
+            parsed_field = {
+                "value": value or "",
+                "rect": iterable_items_to_ints(raw_rect),
+                "label": label or None,
+            }
+            record_fields[name] = parsed_field
 
-        self.ctx.active_pdf_document.fields = record_fields or None
+        self.active_pdf_document.fields = record_fields or None
         return record_fields
 
     @keyword
@@ -563,7 +598,13 @@ class ModelKeywords(LibraryContext):
     ) -> None:
         """Set value for field with given name on the active document.
 
-        Tries to match on field identifier and its label.
+        Tries to match with field's identifier directly or its label.
+
+        :param field_name: Field to update.
+        :param value: New value for the field.
+        :param source_path: Source PDF file path.
+        :raises ValueError: When field can't be found or more than one field matches
+            the given `field_name`.
 
         **Examples**
 
@@ -571,10 +612,6 @@ class ModelKeywords(LibraryContext):
 
         .. code-block:: robotframework
 
-            ***Settings***
-            Library    RPA.PDF
-
-            ***Tasks***
             Example Keyword
                 Open PDF    ./tmp/sample.pdf
                 Set Field Value    phone_nr    077123123
@@ -592,12 +629,6 @@ class ModelKeywords(LibraryContext):
                 pdf.open_pdf("./tmp/sample.pdf")
                 pdf.set_field_value("phone_nr", "077123123")
                 pdf.save_field_values(output_path="./tmp/output.pdf")
-
-        :param field_name: field to update.
-        :param value: new value for the field.
-        :param source_path: source PDF filepath.
-        :raises ValueError: when field can't be found or more than 1 field matches
-            the given `field_name`.
         """
         fields = self.get_input_fields(source_path=source_path)
         if not fields:
@@ -608,11 +639,11 @@ class ModelKeywords(LibraryContext):
         else:
             label_matches = 0
             field_key = None
-            for k, _ in fields.items():
+            for key in fields.keys():
                 # pylint: disable=E1136
-                if fields[k]["label"] == field_name:
+                if fields[key]["label"] == field_name:
                     label_matches += 1
-                    field_key = k
+                    field_key = key
             if label_matches == 1:
                 fields[field_key]["value"] = value  # pylint: disable=E1136
             elif label_matches > 1:
@@ -636,16 +667,20 @@ class ModelKeywords(LibraryContext):
     ) -> None:
         """Save field values in PDF if it has fields.
 
+        :param source_path: Source PDF with fields to update.
+        :param output_path: Updated target PDF.
+        :param newvals: New values when updating many at once.
+        :param use_appearances_writer: For some PDF documents the updated
+            fields won't be visible, try to set this to `True` if you
+            encounter problems. (viewing the output PDF in browser might display the
+            field values then)
+
         **Examples**
 
         **Robot Framework**
 
         .. code-block:: robotframework
 
-            ***Settings***
-            Library    RPA.PDF
-
-            ***Tasks***
             Example Keyword
                 Open PDF    ./tmp/sample.pdf
                 Set Field Value    phone_nr    077123123
@@ -679,22 +714,14 @@ class ModelKeywords(LibraryContext):
                     output_path="./tmp/output.pdf",
                     newvals=new_fields
                 )
-
-        :param source_path: source PDF with fields to update
-        :param output_path: updated target PDF
-        :param newvals: new values when updating many at once
-        :param use_appearances_writer: for some PDF documents the updated
-            fields won't be visible, try to set this to `True` if you
-            encounter problems
         """
         # NOTE:
         # The resulting PDF will be a mutated version of the original PDF,
         # and it won't necessarily show correctly in all document viewers.
         # It also won't show anymore as having fields at all.
         # The tests will XFAIL for the time being.
-
         self.ctx.switch_to_pdf(source_path)
-        reader = self.ctx.active_pdf_document.reader
+        reader = self.active_pdf_document.reader
         if "/AcroForm" in reader.trailer["/Root"]:
             reader.trailer["/Root"]["/AcroForm"].update(
                 {
@@ -703,18 +730,18 @@ class ModelKeywords(LibraryContext):
                     ): PyPDF2.generic.BooleanObject(True)
                 }
             )
-
         writer = PyPDF2.PdfFileWriter()
         if use_appearances_writer:
             writer = self._set_need_appearances_writer(writer)
+
         if newvals:
             self.logger.debug("Updating form fields with provided values for all pages")
             updated_fields = newvals
-        elif self.ctx.active_pdf_document.fields:
+        elif self.active_pdf_document.fields:
             self.logger.debug("Updating form fields with PDF values for all pages")
             updated_fields = {
                 k: v["value"] or ""
-                for (k, v) in self.ctx.active_pdf_document.fields.items()
+                for (k, v) in self.active_pdf_document.fields.items()
             }
         else:
             self.logger.debug("No values available for updating the form fields")
@@ -730,7 +757,7 @@ class ModelKeywords(LibraryContext):
             writer.addPage(page)
 
         if output_path is None:
-            output_path = self.ctx.active_pdf_document.path
+            output_path = self.active_pdf_document.path
         with open(output_path, "wb") as stream:
             writer.write(stream)
 
